import unittest
from test.dummy_model import DummyModel
from tuningdeap import TuningBayes
import json
import os
import logging
import pandas as pd

logging.basicConfig(level=logging.INFO)

class TestWarmStartInitialization(unittest.TestCase):
    def setUp(self):
        cwd = os.getcwd()
        with open(os.path.join(cwd, "test", "tuning_config_test.json"), "r") as file:
            self.tuning_config = json.load(file)
        with open(os.path.join(cwd, "test", "real_test_config.json"), "r") as file:
            self.model_config = json.load(file)

        self.warm_start_path = os.path.join(cwd, "test", "previous_population.csv")
        self.warm_start = pd.read_csv(self.warm_start_path, header=0, index_col=0)

    def test_init_runs_integration(self):
        def eval_function(real_config_updated):
            raise Exception("Should never call this")
        self.n_calls = 0
        tune = TuningBayes(eval_function, self.tuning_config, self.model_config, warm_start_path=self.warm_start_path, n_calls=self.n_calls)
        scores, configs = tune.init_warm_start()
        assert len(configs) == self.n_calls + self.warm_start.shape[0], "did not generate 1 + previously generated: got {} but expected {}".format(len(configs), self.n_calls + self.warm_start.shape[0])



import typing

class DummyModel:
    def __init__(self):
        pass

    def predict(self, real_config_updated: dict) -> float:
        """
        A basic evaluation function that takes in a config file
        :param real_config_updated: the "real" like config file
        :return the scores as a float
        """
        value = real_config_updated["name1"] * real_config_updated["name3"] if real_config_updated["name2"] else 0
        return value

    def predict_no_config(self, values: typing.List) -> float:
        """
        A basic evaluation function with no config file
        :param values: the values from the genetic algorithm
        :return the scores as a float
        """
        return values[0] * 2 if values[1] is True else 1 / (values[2] + 0.001)


import copy
import logging
import math
import os
import random
import typing
import time

from deap import algorithms, base, creator, tools
import numpy as np
import pandas as pd
import warnings

from tuningdeap.base_tuner import TuningBase
from tuningdeap.config_mapping import set_by_path, get_paths

logger = logging.getLogger(__name__)


class TuningDeap(TuningBase):
    """
    A class to genetically tune algorithms. Can handle a config file. Wraps the `deap` package.
    """

    def __init__(
        self, evaluate_outside_function: typing.Callable, tuning_config: dict, original_config: dict = None,
        output_dir: str = None, verbose: bool = False, minimize: bool = True, timeout: float = float("inf"),
        n_generations: int = None, population_size: int = None, random_seed: int = None, warm_start_path: str = None
    ):
        """
        Sets up the class, creates the config mapper if needed, the evaluate function, and the genetic algorithm parameters
        :param evaluate_outside_function: the function to evaluate for the outside caller
        :param tuning_config: the config for the genetic algorithm
        :param original_config: the original_configuration file, if needed, so that we can evaluate
        :param output_dir: the location to write the output files
        :param verbose: whether to report progress to logging while running
        :param minimize: if True, minimize evaluate_outside_function; else maximize it
        :param timeout: the time in seconds that tuningDEAP should continue to generate populations
        :param n_generations: the number of generations to tune
        :param population_size: the number of individuals to evaluate each generation
        :param random_seed: set the random seed for deap
        :param warm_start_path: the path to a csv to warm start with
        """
        super().__init__(evaluate_outside_function, tuning_config, original_config, output_dir=output_dir, verbose=verbose, minimize=minimize, timeout=timeout, 
                        random_seed=random_seed, warm_start_path=warm_start_path)
        self.population_size = population_size
        self.n_generations = n_generations

        # override the base class and add the tuple that deap needs
        self.evaluate_function = self.create_evaluate_func_deap()

        self.validate_config()
        self.toolbox = base.Toolbox()
        self._instantiate_attributes()

        if self.verbose:
            logger.info("Finished initializing.")

    def create_evaluate_func_deap(self) -> typing.Callable:
        """
        The function to create the function used by the genetic algorithm.
        We have to wrap the evaluation function with another function to change the chromosomes into a dict like the `original_config` file.
        """
        def evaluate_function(values: typing.List) -> typing.Tuple:
            """
            Simply map the chromosome values into a dict like the original config file and return that to the evaluate function
            :param values: the "chromosone" or parameter for this "individual"
            :return a tuple containing the score of fitness
            """
            config = self.map_tuning_config_back(values)
            try:
                return (self.evaluate_outside_function(config), )
            except Exception as e:
                warnings.warn("there was an error while evaluating the function: {}".format(e))
                return tuple((float('inf'), )) if self.minimize else tuple((float("-inf"), ))
        return evaluate_function

    def run(self) -> (typing.List, float):
        """
        The main function to run the evoluationary algorithm
        """
        population = self.toolbox.population(n=self.population_size)
        hof = tools.HallOfFame(1)

        if self.warm_start_path is not None:
            population = self.create_population_from_csv(population)

        if self.output_dir is not None:
            halloffame = tools.HallOfFame(maxsize=self.population_size)
        topone = tools.HallOfFame(maxsize=1)

        ending_time = time.time() + self.timeout
        self.gen = 0
        while self.gen < self.n_generations and time.time() < ending_time:
            self.gen += 1
            if self.verbose:
                print("On generation {}".format(self.gen))

            # create and validate new generation by geneticism
            final_population = self.get_new_generation(population)
            
            # Evaluate the individuals with an invalid fitness
            population = final_population
            invalid_ind = [ind for ind in population if not ind.fitness.valid]
            fitnesses = self.toolbox.map(self.toolbox.evaluate, invalid_ind)
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit

            if self.output_dir is not None:
                halloffame.update(population)
                # it's a list of lists
                individuals = halloffame.items
                # reverse halloffame items since they did it reverse
                individuals.reverse()
                scores = [item.wvalues[0] * self.optimize for item in halloffame.keys]
                if self.verbose:
                    print("The results from generation {} are:".format(self.gen))
                self.create_dataset_from_results(individuals, scores, self.output_dir is not None, self.verbose, prefix="generation-{}".format(self.gen))
                halloffame.clear()

            if len(population) != 0:
                topone.update(population)
            population = self.toolbox.select(population, k=len(population))

        if len(topone) == 0:
            return [], float("inf") if self.minimize else float("-inf")

        return topone.items[0], topone.keys[0].wvalues[0] * self.optimize

    def get_new_generation(self, population: list) -> list:
        """
        This function gets the new generation of the population, enforces, and verfies the limits and size
        :param population: the list of individuals, each a list
        :return the final population list of lists
        """
        # enforce our own limits on parameters regardless of mutations
        iterations = 0
        final_population = []
        while len(final_population) < self.population_size and iterations < 10:
            iterations += 1
            init_population = algorithms.varAnd(population, self.toolbox, cxpb=0.5, mutpb=0.2)
            final_population += self.enforce_limits(init_population)
            # if too big, scale it down
            final_population = final_population[:self.population_size]
        
        # verify population
        small_population = len(final_population) < self.population_size
        if iterations == 10 and small_population:
            raise EnvironmentError("bounds could not be enforced or population too small")
        assert small_population == False, "population was smaller than it should have been"
        assert len(population) == len(final_population), "populations are not the same size. Expected {} but got {}".format(len(population), len(final_population))
        return final_population

    def validate_config(self):
        """
        Used to validate essential config properties
        """
        if self.population_size is None:
            assert "population_size" in self.tuning_config, "Tuning config did not have population size!"
            self.population_size = self.tuning_config["population_size"]
        
        if self.n_generations is None:
            assert "n_generations" in self.tuning_config, "Tuning config did not have n_generations!"
            self.n_generations = self.tuning_config["n_generations"]

    def getConfigOr(self, key: str, alternate: typing.Callable) -> typing.Callable:
        """
        Used to override genetic algorithm parameters
        :param key: the key to use to lookup the function
        :param alternate: the default function to use if the key does not exist
        :return the function to be used
        """
        if key in self.tuning_config:
            # do something
            # TODO: make this dynamicly change based on input
            pass
        return alternate

    def _instantiate_attributes(self):
        """
        This function sets the necessary class values and creates the layout for the genetic algorithm
        """
        # set up the general genetic algorithm functions
        self.toolbox.register("evaluate", self.evaluate_function)
        self.toolbox.register("mate", self.getConfigOr("mate", tools.cxTwoPoint))
        self.toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
        self.toolbox.register("select", self.getConfigOr("select", tools.selTournament), tournsize=3)
        # 1.0 means that we are maximizing the function
        self.optimize = -1 if self.minimize else 1
        creator.create("FitnessMax", base.Fitness, weights=(self.optimize,))
        creator.create("Individual", list, fitness=creator.FitnessMax)

        chromosome = []
        self.bool_values = []
        self.categorical_values = {}
        self.order_of_keys = []
        # for each attribute given to model, set up the required limits and parameters for the genetic algorithm
        for attribute_name, attribute_map in self.tuning_config["attributes"].items():
            # gather the needed info
            self.order_of_keys.append(attribute_name)
            param_type = attribute_map["type"]
            if param_type not in ["bool", "categorical"]:
                min_val = attribute_map["min"]
                max_val = attribute_map["max"]
                step_size =  attribute_map["step"] if "step" in attribute_map else 1
            # create the values in the program
            if param_type == "int":
                self.toolbox.register("attr_int", random.randrange, min_val, max_val, step_size)
                chromosome.append(self.toolbox.attr_int)
            if param_type == "bool":
                self.toolbox.register("attr_bool", random.randint, 0, 1)
                chromosome.append(self.toolbox.attr_bool)
                self.bool_values.append(attribute_name)
            if param_type == "float":
                self.toolbox.register("attr_float", self.rand_with_step, min_val, max_val, step_size)
                chromosome.append(self.toolbox.attr_float)
            if param_type == "categorical":
                self.toolbox.register("attr_int", random.randrange, 0, len(attribute_map["values"]), 1)
                chromosome.append(self.toolbox.attr_int)
                self.categorical_values[attribute_name] = attribute_map["values"]

        assert len(chromosome) != 0, "No values were added for the genetic algorithm"
        # finalize setting up the algorithm by specifying the chromosones, individual makeup, and the population
        self.toolbox.register("individual", tools.initCycle, creator.Individual, chromosome, n=1)
        self.toolbox.register("population", tools.initRepeat, list, self.toolbox.individual)

    def create_population_from_csv(self, init_population: typing.List[typing.List]):
        """
        This function is used to "warm-start" or initialize a population from previous values
        :param init_population: the deap generated list of individuals
        :return the init_population that has been populated by the csv file
        """
        previous_population = self.read_and_validate_previous(self.warm_start_path)

        # get the best given if more than population_size given
        previous_population.sort_values(by=["score"])
        starting_population = previous_population.iloc[:self.population_size, ].to_numpy().tolist()

        # fill in any missing inidivduals from the randomly created population
        if len(starting_population) < self.population_size:
            starting_population += init_population[:self.population_size - len(starting_population)]

        assert len(starting_population) == self.population_size, "could not create a population size of {}".format(self.population_size)
        return starting_population



import unittest
from tuningdeap import TuningDeap
import logging

logging.basicConfig(level=logging.INFO)


class TestTiming(unittest.TestCase):

    def setUp(self):
        self.tuning_config = {
            "population_size": 1,
            "n_generations": 1,
            "attributes": {
                "name1": {
                    "type": "float",
                    "min": 0,
                    "max": 1,
                    "step": 0.1
                },
                "name2": {
                    "type": "bool"
                },
                "name3": {
                    "type": "int",
                    "min": 1,
                    "max": 5,
                    "step": 1
                },
                "name4": {
                    "type": "categorical",
                    "values": ["a", "b", "c"]
                }
            }
        }

    def test_population_and_generations_from_config(self):
        def eval_function(chromosomes):
            return 1
        tune = TuningDeap(eval_function, self.tuning_config, verbose=False)
        assert tune.n_generations == self.tuning_config["n_generations"], "did not read n_generations from the config file like it should have"
        assert tune.population_size == self.tuning_config["population_size"], "did not read n_generations from the config file like it should have"
    
    def test_population_and_generations_from_constructor(self):
        def eval_function(chromosomes):
            return 1
        n_generations = 10
        population_size = 10
        tune = TuningDeap(eval_function, self.tuning_config, verbose=False, population_size=population_size, n_generations=n_generations)
        assert tune.n_generations == n_generations, "did not read n_generations from the constructor like it should have"
        assert tune.population_size == population_size, "did not read n_generations from the constructor like it should have"
    


import pandas as pd
import numpy as np
import skopt
import copy
import logging
import math
import os
import random
import typing
import time
import warnings

from tuningdeap.base_tuner import TuningBase
from tuningdeap.config_mapping import set_by_path, get_paths

from skopt.space import Real, Integer
from skopt.utils import use_named_args
from skopt import gp_minimize

logger = logging.getLogger(__name__)

class TuningBayes(TuningBase):
    """
    A class to tune algorithms with bayesian optimization. Can handle a config file. Wraps the `sckit-optimize` package.
    """

    def __init__(
        self, evaluate_outside_function: typing.Callable, tuning_config: dict, original_config: dict = None,
        output_dir: str = None, verbose: bool = False, minimize: bool = True, timeout: float = float("inf"),
        n_calls: int = None, n_random_starts: int = 1, random_seed: int = None, jobs: int = None, 
        n_points: int = None, warm_start_path: str = None
    ):
        """
        Sets up the class, creates the config mapper if needed, the evaluate function, and the genetic algorithm parameters
        :param evaluate_outside_function: the function to evaluate for the outside caller
        :param tuning_config: the config for the genetic algorithm
        :param original_config: the original_configuration file, if needed, so that we can evaluate
        :param output_dir: the location to write the output files
        :param verbose: whether to report progress to logging while running
        :param minimize: if True, minimize evaluate_outside_function; else maximize it
        :param timeout: the time in seconds that tuningDEAP should continue to generate populations
        :param n_calls: the number of calls to make
        :param n_random_starts: the number of times to randomly start
        :param random_seed: set the random seed
        :param jobs: the number of threads to run
        :param n_points: the number of points to run simultaneously
        :param warm_start_path: the path to a csv to warm start with
        """
        super().__init__(evaluate_outside_function, tuning_config, original_config, output_dir=output_dir, verbose=verbose, minimize=minimize, timeout=timeout, 
                        random_seed=random_seed, warm_start_path=warm_start_path)
        self.n_random_starts = n_random_starts
        self.n_points = n_points
        self.jobs = jobs
        self.n_calls = n_calls
        if self.n_calls == 0:
            warnings.warn("TuningBayes called with 0 calls.  No calls will be made")

        if not self.minimize:
            raise Exception("Bayesian Optimization does not support maximization. Please use tuningdeap or reformat the problem.")

        self.validate_config()
        self.space = self._instantiate_space()

        if self.random_seed is None:
            self.random_seed = random.randint(0, 9999999)
        if self.verbose:
            logger.info("Finished initializing.")

    def run(self):
        config_values = []
        scores = []
        args, no_warm_start_args = self.prepare_args()
        if self.timeout != float("inf"):
            start_time = time.time()
            # do the first so we can plug the values back in and do one iteration each time
            res_gp = gp_minimize(self.evaluate_function, self.space, n_calls=1, random_state=self.random_seed, verbose=self.verbose, **args)
            while time.time() < start_time + self.timeout:
                # this may add some overhead by re-creation, but is the easiest way to do a timed tuning
                res_gp = gp_minimize(self.evaluate_function, self.space, n_calls=1, random_state=self.random_seed, verbose=self.verbose, 
                                x0=res_gp.x_iters, y0=res_gp.func_vals, **no_warm_start_args)
            
            if self.verbose or self.output_dir is not None:
                self.create_dataset_from_results(res_gp.x_iters, res_gp.func_vals, self.output_dir is not None, self.verbose, prefix="bayes")


        else:
            res_gp = gp_minimize(self.evaluate_function, self.space, n_calls=self.n_calls, random_state=self.random_seed, verbose=self.verbose, **args)

            if self.verbose or self.output_dir is not None:
                self.create_dataset_from_results(res_gp.x_iters, res_gp.func_vals, self.output_dir is not None, self.verbose, prefix="bayes")

        # save values
        self.x_iters = res_gp.x_iters
        self.scores = res_gp.func_vals
        return res_gp.x, res_gp.fun
 
    def validate_config(self):
        """
        Used to validate essential config properties
        """
        if self.n_calls is None:
            assert "n_calls" in self.tuning_config, "Tuning config did not have `n_calls`!"
            self.n_calls = self.tuning_config["n_calls"]
        
    def _instantiate_space(self):
        space = []
        self.bool_values = []
        self.categorical_values = {}
        self.order_of_keys = []
        # for each attribute given to model, set up the required limits and parameters for the genetic algorithm
        for attribute_name, attribute_map in self.tuning_config["attributes"].items():
            # gather the needed info
            self.order_of_keys.append(attribute_name)
            param_type = attribute_map["type"]
            if param_type not in ["bool", "categorical"]:
                min_val = attribute_map["min"]
                max_val = attribute_map["max"]
                step_size =  attribute_map["step"] if "step" in attribute_map else 1
                if step_size != 1:
                    warnings.warn("TuningBayes can not take a step value: ignoring...")
            # create the values in the program
            if param_type == "int":
                space.append(Integer(min_val, max_val, name=attribute_name))
            if param_type == "bool":
                space.append(Integer(0, 1, name=attribute_name))
                self.bool_values.append(attribute_name)
            if param_type == "float":
                space.append(Real(min_val, max_val, name=attribute_name))
            if param_type == "categorical":
                space.append(Integer(0, len(attribute_map["values"]) - 1, name=attribute_name))
                self.categorical_values[attribute_name] = attribute_map["values"]

        assert len(space) != 0, "No values were added for the bayesian optimization"
        return space

    def prepare_args(self):
        """
        Used to prepare the arguments needed from the bayesian tuner
        :returns a dict representing the function parameters and values, an identical one not containing warm start values
        """
        args = {}
        if self.jobs is not None:
            args["jobs"] = self.jobs
        if self.n_points is not None:
            args["n_points"] = self.n_points
        if self.n_random_starts is not None:
            args["n_random_starts"] = self.n_random_starts
        if self.warm_start_path is not None:
            scores, config_values = self.init_warm_start()
            warm_start_args = args.copy()
            warm_start_args["x0"] = config_values
            warm_start_args["y0"] = scores
        else:
            warm_start_args = args
        return warm_start_args, args

    def init_warm_start(self):
        """
        Used for validating and preparing previous runs for a warm start
        :returns 
            scores: the scores from the previous warm start
            config_values: the parameters that generated those scores
        """
        warm_start = self.read_and_validate_previous(self.warm_start_path)
        scores = warm_start["score"].tolist()
        config_values = warm_start.drop("score", axis=1).to_numpy().tolist()
        return scores, config_values



import unittest
from test.dummy_model import DummyModel
from tuningdeap import TuningDeap
import json
import os
import logging

logging.basicConfig(level=logging.INFO)

class TestStartingPopulationInitialization(unittest.TestCase):
    def setUp(self):
        cwd = os.getcwd()
        with open(os.path.join(cwd, "test", "tuning_config_test.json"), "r") as file:
            self.tuning_config = json.load(file)
        with open(os.path.join(cwd, "test", "real_test_config.json"), "r") as file:
            self.model_config = json.load(file)

        # an example of a randomly generated population for the tuning_config
        self.random_generation = [
            [0.5, 1, 2, 1],
            [0.30000000000000004, 1, 3, 1],
            [0.4, 1, 1, 1], [0.2, 0, 3, 1],
            [0.4, 0, 1, 0],
            [0.1, 1, 3, 0],
            [0.1, 1, 3, 1],
            [0.0, 0, 2, 2],
            [0.30000000000000004, 1, 2, 2],
            [0.4, 1, 1, 1]
        ]
        self.warm_start_path = os.path.join(cwd, "test", "previous_population.csv")

    def test_init_population_integration(self):
        def eval_function(real_config_updated):
            return DummyModel().predict(real_config_updated)
        tune = TuningDeap(eval_function, self.tuning_config, self.model_config, warm_start_path=self.warm_start_path)
        # change the generation number so we get the initial population
        warm_started_population = tune.create_population_from_csv(self.random_generation)
        assert warm_started_population != self.random_generation, "did not generate a new population from the csv file"



import unittest
import logging
from examples.tune_basic_mlp import execute_example_tuning

logging.basicConfig(level=logging.INFO)

class TestFullMLP(unittest.TestCase):

    def test_example_run_full_mlp(self):
        # make sure the example runs successfully
        execute_example_tuning()



import unittest
from tuningdeap import TuningDeap
import json
import logging
import numpy as np
import pandas as pd

logging.basicConfig(level=logging.INFO)

class TestPopulationSize(unittest.TestCase):

    def setUp(self):
        self.tuning_config = {
            "population_size": 100,
            "attributes": {
                "name1": {
                    "type": "float",
                    "min": 0,
                    "max": 1,
                    "step": 0.1
                },
                "name2": {
                    "type": "bool"
                },
                "name3": {
                    "type": "int",
                    "min": 1,
                    "max": 5,
                    "step": 1
                },
                "name4": {
                    "type": "categorical",
                    "values": ["a", "b", "c"]
                }
            }
        }

    def test_large_population_size(self):
        def eval_function(chromosomes):
            return 1
        # run and see if it hits the assertion for too small of a population
        tune = TuningDeap(eval_function, self.tuning_config, n_generations=10, population_size=10000)
        _, _ = tune.run()

    def test_population_output_mixup(self):
        def eval_function(chromosomes):
            return np.sum(chromosomes[:len(chromosomes) - 1])

        n_generations = 10
        tune = TuningDeap(eval_function, self.tuning_config, minimize=False, output_dir="./tmp", n_generations=n_generations)
        _, _ = tune.run()
        # check the generations
        for generations in range(n_generations):
            gen = pd.read_csv("./tmp/generation-{}.csv".format(generations + 1), header=0, index_col=0)
            for index, row in gen.iterrows():
                assert eval_function(row[:len(row) - 1].to_numpy()) == row[-1], "did not get the score expected: output: {}, expected: {}".format(eval_function(row[:len(row) - 1].to_numpy()), row[-1])




import copy
import logging
import math
import os
import random
import typing
import time

from deap import algorithms, base, creator, tools
import numpy as np
import pandas as pd
import warnings

from tuningdeap.config_mapping import set_by_path, get_paths

logger = logging.getLogger(__name__)


class TuningBase:
    """
    A base class for tuning algorithms. Can handle a config file.
    """

    def __init__(
        self, evaluate_outside_function: typing.Callable, tuning_config: dict, original_config: dict = None, warm_start_path: str = None,
        output_dir: str = None, verbose: bool = False, minimize: bool = True, timeout: float = float("inf"), random_seed: int = None
    ):
        """
        Sets up the class, creates the config mapper if needed, the evaluate function, and the genetic algorithm parameters
        :param evaluate_outside_function: the function to evaluate for the outside caller
        :param tuning_config: the config for the genetic algorithm
        :param original_config: the original_configuration file, if needed, so that we can evaluate
        :param output_dir: the location to write the output files
        :param verbose: whether to report progress to logging while running
        :param minimize: if True, minimize evaluate_outside_function; else maximize it
        :param timeout: the time in seconds that tuningDEAP should continue to generate populations
        :param n_generations: the number of generations to tune
        :param population_size: the number of individuals to evaluate each generation
        :param warm_start_path: the string path to a csv file containing previous tuned values and scores
        :param random_seed: set the random seed for deap
        """
        self.tuning_config: dict = tuning_config
        self.using_config = False
        self.warm_start_path = warm_start_path

        self.output_dir = output_dir
        if self.output_dir is not None and not os.path.isdir(self.output_dir):
            os.mkdir(self.output_dir)

        self.minimize = minimize
        self.timeout = timeout
       
        self.random_seed = random_seed
        if self.random_seed is not None:
            random.seed(self.random_seed)

        self.verbose = verbose
        if self.verbose:
            logger.setLevel(logging.INFO)
            logger.info("verbose is set to: {}".format(self.verbose))

        if original_config is not None:
            self.original_config: dict = original_config
            self.using_config = True
            self.map_config: dict = self.map_tuning_config_to_original()

        
        self.evaluate_outside_function = evaluate_outside_function
        self.evaluate_function: typing.Callable = self.create_evaluate_func()

    def create_evaluate_func(self) -> typing.Callable:
        """
        The function to create the function used by the tuning algorithm.
        We have to wrap the evaluation function with another function to change the parameters into a dict like the `original_config` file.
        """
        def evaluate_function(values: typing.List) -> float:
            """
            Simply map the chromosome values into a dict like the original config file and return that to the evaluate function
            :param values: the "chromosone" or parameter for this "individual"
            :return a scalar containing the score
            """
            config = self.map_tuning_config_back(values)
            try:
                return self.evaluate_outside_function(config)
            except Exception as e:
                raise Exception("there was an error while evaluating the function: {}".format(e))
                return float('inf') if self.minimize else float("-inf")
        return evaluate_function

    def enforce_limits(self, init_population: typing.List[typing.List]) -> typing.List[typing.List]:
        """
        Enforces the given limits on the mutated/crossovered population since deap does not.
        :param init_population: a list of lists containing `inividual's chromosomes`
        :return a list of lists representing the valid members of the population
        """
        final_population = []
        for individual in init_population:
            invalid = False
            for index, parameter in enumerate(individual):
                parameter_info = self.tuning_config["attributes"]
                parameter_name = self.order_of_keys[index]
                parameter_map = parameter_info[parameter_name]
                attribute_type = parameter_map["type"]

                if attribute_type not in ["bool", "categorical"]:
                    min_val = parameter_map["min"]
                    max_val = parameter_map["max"]
                    step_size =  parameter_map["step"] if "step" in parameter_map else 1
                    if min_val <= parameter <= max_val:
                        try:
                            # see if strict bounds are enforced
                            if "strict" in parameter_info and self.enforce_steps(parameter, min_val, max_val, step_size):
                                continue
                        except IndexError:
                            continue
                    else:
                        if self.verbose:
                            logger.info("Rejecting float/int parameter for individual: {}={} with bounds: {}".format(parameter_name, parameter,
                                        " ".join([str(item) for item in [min_val, max_val, step_size]])))
                        invalid = True
                        break

                elif attribute_type == "bool":
                    # handle the boolean case
                    if parameter == 0 or parameter == 1:
                        continue
                    else: 
                        if self.verbose:
                            logger.info("Rejecting boolean parameter for individual: {} with bounds: {}".format(parameter_name, parameter))
                        invalid = True
                        break

                elif attribute_type == "categorical":
                    values = parameter_map["values"]
                    if 0 <= parameter <= len(values):
                        continue
                    else:
                        if self.verbose:
                            logger.info("Rejecting categorical parameter for individual: {}={} with bounds: 0-{}".format(parameter_name, parameter,
                                        " ".join(len(values))))
                        invalid = True
                        break
                else:
                    raise Exception("Wrong type of value for parameter: {}".format(attribute_type))

            if not invalid:
                final_population.append(copy.deepcopy(individual))

        if self.verbose:
            logger.info("Enforcing limits for the population. Have {} after rejecting the invalid: population_size is {}".format(len(final_population), self.population_size))
        return final_population

    def enforce_steps(self, parameter, min_val: int, max_val: int, step_size) -> bool:
        # TODO: add the enforcement for this step
        return True

    def rand_with_step(self, low: float, high: float, step: float, count: int = 1, bias: bool = True):
        """
        A function to return a random float with steps
        :param low: the `low` step to start out on
        :parm high: the `high` end of the range of options
        :param step: the step size to take between `low` and `high`
        :param count: how many floats to return
        :param bias: whether to start on the `low` step
        :return a float between `low` and `high`, with steps sizes of size `step`
        """
        n = 1 / step
        if count > 1:
            val = np.random.randint(low * n, high * n, count) * step
        else:
            val= np.random.randint(low * n,high * n) * step

        if bias:
            bias = math.ceil(low / step) * step - low
        else:
            bias = 0
        return val - bias

    def map_tuning_config_to_original(self):
        """
        The function that maps the tuning config and genetic algorithm parameters to the original config for evaluation
        """
        paths = get_paths(self.original_config)
        mapping = {}
        for key, value in self.tuning_config["attributes"].items():
            for index, path in enumerate(paths):
                if key == path[-1]:
                    mapping[key] = paths[index]
                    continue
            if key not in mapping:
                raise ValueError("Tuning config contained a attribute parameter not in the original config: {}".format(key))
        return mapping

    def map_tuning_config_back(self, chromosome: typing.List) -> typing.Container:
        """
        Maps the list of chromosomes back into a dict with the same structure as the original config file
        :param chromosome: a list of values for that individual in the population
        :return a dict containing the individuals values, but in the original config file form if using a config file, else a list of parameters
        """
        if self.using_config:
            new_config = copy.deepcopy(self.original_config)
            for index, name in enumerate(self.order_of_keys):
                current_value = chromosome[index]
                path_to_original = self.map_config[name]
                set_by_path(new_config, path_to_original, current_value, is_bool=name in self.bool_values,
                            categorical_value=self.categorical_values[name][current_value] if name in self.categorical_values else None)
            return new_config
        else:
            return chromosome

    def create_dataset_from_results(self, params: typing.List[typing.List], scores: typing.List[float], output_to_file: bool = False, 
                                    verbose: bool = False, prefix="bayes"):
        """
        This function is abstracted away to gather the info needed to write the results to file.  It will gather the correct names, scores and 
        deal with outputting or printing them depending on the config
        :param params: the parameters for that hyperparameter run
        :param scores: the scores for each parameter run
        :param output_to_file: a boolean indicating whether or not to write to file
        :param verbose: a boolean indicating whether or not to print the results
        :param prefix: the prefix for the file to be written
        :return the results in a csv format with columns "score" and the names of the parameters
        """
        assert len(params) == len(scores), "scores and params were different lengths! params: {}, scores: {}".format(len(params), len(scores))
        full_named_items = []
        for ind_index, ind_params in enumerate(params):
            named_items = {}
            # get the named parameters
            for index, item in enumerate(ind_params):
                name = self.order_of_keys[index]
                named_items[name] =  item
            # flip the sign if we need to
            named_items["score"] = scores[ind_index]
            full_named_items.append(named_items)
        # write out to file
        results = pd.DataFrame(full_named_items)
        if output_to_file:
            output_path = os.path.join(self.output_dir, '{}.csv'.format(prefix))
            results.to_csv(output_path)
        if verbose:
            print(results)

        return results

    def read_and_validate_previous(self, path):
        """
        A common function for warm starting the tuning.  Takes in a csv filepath and returns the validated DataFrame
        :param path: a str containing the path to the warm start csv file
        :return a Pandas DataFrame of the warm start
        """
        if not os.path.isfile(path):
            raise ValueError("path to initial warm start: {}, was not a valid file or did not exist".format(self.warm_start_path))

        warm_start = pd.read_csv(path, header=0, index_col=False)
        assert set(warm_start.columns.tolist()) == set(self.order_of_keys + ["score"]), "given csv file did not contain the same parameters as the tuning config"
        return warm_start



import unittest
from test.dummy_model import DummyModel
from tuningdeap import TuningDeap
import json
import os
import logging

logging.basicConfig(level=logging.INFO)

class TestRepoducibility(unittest.TestCase):
    def setUp(self):
        cwd = os.getcwd()
        with open(os.path.join(cwd, "test", "tuning_config_test.json"), "r") as file:
            self.tuning_config = json.load(file)
        with open(os.path.join(cwd, "test", "real_test_config.json"), "r") as file:
            self.model_config = json.load(file)


    # def test_same_output(self):
    #     # TODO: fix this test
    #     def eval_function(real_config_updated):
    #         return DummyModel().predict(real_config_updated)
    #     tune1 = TuningDeap(eval_function, self.tuning_config, self.model_config, minimize=False, random_seed=0)
    #     best_config1, best_score1 = tune1.run()

    #     tune2 = TuningDeap(eval_function, self.tuning_config, self.model_config, minimize=False, random_seed=0)
    #     best_config2, best_score2 = tune2.run()
    #     assert best_score1 == best_score2, "did not reproduce the same score: {} vs {}".format(best_score1, best_score2)
    #     assert best_config1 == best_config2, "did not reproduce the same best config: {} vs {}".format(best_config1, best_config2)
